(*  Title:      Lemmas.thy
    Author:     Ludovic Henrio 
                2008

    Note:       contains a few operators and lemmas not particular to any application domain
*)

header {* A few generic constructs and lemmas not particularly related to components*}

theory Lemmas imports ListPre
begin
subsection{*listfromoption*}
primrec ListFromOption:: " 'a option \<Rightarrow> 'a list" where
  "ListFromOption None = []" |
  "ListFromOption (Some a) = [a]"


lemma lengthListFromOption: "length (ListFromOption a)\<le>1"
by (case_tac a,auto);
subsection{*lemmas on Isabelel constructs *}

lemma foldl_and[rule_format]: " \<forall> a . foldl (op \<and>) a l \<longrightarrow> a"
apply (induct_tac l);
apply auto;
done;
(* FORMATTED: lemma foldl_and: foldl op \<and> ?a ?l \<Longrightarrow> ?a *)

lemma foldl_and_rec[rule_format]: " foldl (op \<and>) a l \<longrightarrow>  foldl (op \<and>) True  l"
apply (induct_tac l);
apply clarify;
apply force;
apply simp;
apply clarify;
apply (subgoal_tac a);
apply simp;
apply (subgoal_tac "a\<and>aa");
apply simp;
apply (rule_tac l=list in foldl_and);
apply auto;
done;
(* FORMATTED: lemma foldl_and_rec: foldl op \<and> ?a ?l \<Longrightarrow> foldl op \<and> True ?l *)

lemma cardinal_singleton: "card(S)=1 \<Longrightarrow>\<exists>a. S={a}"
apply (case_tac "finite S");
apply auto;
apply (case_tac "\<exists> a. S={a}");
apply force;
apply (case_tac "S={}");
apply simp+;
apply (subgoal_tac "card S>1");
apply auto;
apply (subgoal_tac "card (S-{x})=0");
apply (drule_tac x=x in spec);
apply (subgoal_tac "finite (S-{x})")
apply (drule_tac A="S-{x}" in card_0_eq);
apply auto;
done;  

subsubsection{*list lemmas*}
lemma list_inter_distinct[rule_format]: " \<forall> L2. ((list_inter L1 L2 = [] \<and> distinct L1 \<and> distinct L2) = distinct (L1@L2))"
apply (induct_tac L1);
apply auto;
done
lemma list_inter_distinctimp: " \<lbrakk>list_inter L1 L2 = [];  distinct L1;  distinct L2 \<rbrakk> \<Longrightarrow> distinct (L1@L2)"
apply (insert list_inter_distinct [of L1 L2], auto);
done
lemma list_inter_distinctrev: "distinct (L1@L2) \<Longrightarrow>(list_inter L1 L2 = [] \<and> distinct L1 \<and> distinct L2)"
apply (insert list_inter_distinct [of L1 L2], auto);
done

lemma list_inter_disjoint[rule_format]: " \<forall> L2. list_inter L1 L2 = [] = (set L1 \<inter> set L2 ={})"
apply (induct_tac L1);
apply auto;
done;

lemma listlistith_rec: " \<forall> i1 i2 j1 j2. (i1<length L \<and>j1<length (L!i1) \<and>
                    i2<length L \<and>j2<length (L!i2) \<and>i1\<noteq>i2\<longrightarrow>
(((listsum (map length (take i1 L)))+j1)\<noteq>((listsum (map length (take i2 L))))+j2))"
apply (induct_tac L);
apply force;
apply clarify;
apply (case_tac i1);
apply (case_tac i2);
apply simp;
apply (simp add:  nth_append);
apply (case_tac i2);
apply force;
apply (drule_tac x=nat in spec)
apply (drule_tac x=nata in spec)
apply (drule_tac x=j1 in spec)
apply (drule_tac x=j2 in spec)
apply auto;
done;
lemma listlistith: "\<lbrakk>i\<noteq>i';i<length L ;j<length (L!i) ;i'<length L ;k<length (L!i')\<rbrakk>
 \<Longrightarrow>listsum (map length (take i L))+j\<noteq>listsum (map length (take i' L))+k"
apply (insert listlistith_rec [of L])
apply auto;
done;

lemma listtake_length_rec: "\<forall> i m. (i<length L \<and> m< length (L!i)) \<longrightarrow> (listsum (map length (take i L)))+m<listsum (map length L)"
apply (induct_tac L);
apply force;
apply clarify;
apply (case_tac i);
apply (simp add:  nth_append);
apply force;
done;

lemma listtake_length: "\<lbrakk>i<length L; m< length (L!i)\<rbrakk> \<Longrightarrow> (listsum (map length (take i L)))+m<listsum (map length L)"
by (insert listtake_length_rec, force)


lemma foldUn[rule_format]: " \<forall> A . x\<in> foldl (\<lambda>x y. x\<union>(f y)) A B = (x\<in>A \<or> (\<exists> b \<in> (set B). x\<in>f b))"
apply (induct_tac B);
apply auto;
done;

(*FORMATED  foldUn:
    ?x \<in> foldl (\<lambda>x y. x \<union> ?f y) ?A ?B \<Longrightarrow> ?x \<in> ?A \<or> (\<exists>b\<in>set ?B. ?x \<in> ?f b) *)


lemma disjoint_diff: "A\<inter>B={} \<Longrightarrow> (A-C)\<inter>B={}"
apply force;
done;


lemma map_remove1[rule_format]: "distinct (map f l) \<longrightarrow> distinct (map f (remove1 x l))"
apply (induct_tac l);
apply auto;
apply (subgoal_tac "xa\<in> set list");
apply auto;
done;
end