
header {* GCM Component  Reconfiduration constructe and lemmas *}


theory Reconfig imports CpList
begin

section{*Completeness*}

constdefs allExternalItfsBound:: "Component \<Rightarrow>Binding set \<Rightarrow> bool" 
  "allExternalItfsBound C bs ==
 \<forall> IN itype. (getItfs C)(IN) = Some itype \<longrightarrow> (case (kind itype) of
    Client => \<exists> b\<in>bs. src b = ((getName C).\<^sub>iIN) |
    Server => \<exists> b\<in>bs. dest b = ((getName C).\<^sub>iIN))"

constdefs allInternalItfsBound:: "Component \<Rightarrow>Binding set \<Rightarrow> bool" 
  "allInternalItfsBound C bs ==
 \<forall> IN itype. (getItfs C)(IN) = Some itype \<longrightarrow> (case (kind itype) of
    Client => \<exists> b\<in>bs. dest b = (This IN) |
    Server => \<exists> b\<in>bs. src b = (This IN)) "

declare ExternalToInternal_def[simp];


primrec Complete::"Component\<Rightarrow>bool" and
        CompleteList::"Component list\<Rightarrow>bool" where
  "Complete (Primitive N itf s) = True" |
  "Complete (Composite N itf sub bindings s) = (
         (\<forall> C\<in>set sub. allExternalItfsBound C bindings) \<and> (* all subcomponents bound*)
         (allInternalItfsBound (Composite N itf sub bindings s) bindings) \<and>
                                       (* all internal interfaces bound*)
         (CompleteList sub) (* recursion *) )" |
  "CompleteList [] =  True" |
  "CompleteList (C#CL) = (Complete C \<and> CompleteList CL)"


section{*unbind primitive*}

primrec unbind:: "Component\<Rightarrow>Binding\<Rightarrow>Component" where
  "unbind (Primitive N itf s) b = (Primitive N itf s)" |
  "unbind (Composite N itf sub bindings s) b = (Composite N itf sub (bindings - {b}) s)"

lemma unbinding_incomplete[rule_format]:
"b\<in>bindings \<longrightarrow>  CorrectComponentStructure (Composite N itf sub bindings s)
\<longrightarrow> \<not> Complete (unbind (Composite N itf sub bindings s) b)"
apply (auto);
apply (case_tac "src b");
apply (simp add: allInternalItfsBound_def)
apply (drule_tac x=b in bspec);
apply assumption;
apply simp+;
apply (elim conjE exE)
apply (drule_tac x=Name in spec);
apply simp+;
apply (elim bexE)
apply simp+;
apply (simp add: NoDuplicateSrc_def);
apply (drule_tac x="ba" in bspec);
apply simp;
apply (drule_tac x="b " in bspec);
apply simp+;
apply (simp add: allExternalItfsBound_def)
apply (drule_tac x=b in bspec);
apply assumption;
apply simp+;
apply (elim conjE exE)
apply (case_tac "sub^Name1");
apply simp+;
apply (drule_tac x=a in bspec);
apply (rule getCp_inlist2);
apply simp+;
apply (drule_tac x=Name2 in spec);
apply simp+;
apply (elim bexE);
apply (simp add: NoDuplicateSrc_def);
apply (drule_tac x="ba" in bspec);
apply simp;
apply (drule_tac x="b " in bspec);
apply (simp add: getCp_inlist)+;
done;

section{*Replace*}

primrec RenameQName:: "QualifiedName\<Rightarrow>Name \<Rightarrow>Name \<Rightarrow> QualifiedName" where 
"RenameQName (This IN) oldName newName = This IN" |
"RenameQName (CN.\<^sub>iIN)  oldName newName = ((if (CN = oldName) then newName else CN).\<^sub>iIN)"

constdefs RenameBinding:: "Binding\<Rightarrow>Name \<Rightarrow>Name \<Rightarrow> Binding"
"RenameBinding b oldName newName == 
\<lparr>src = RenameQName (src b) oldName newName, dest = RenameQName (dest b) oldName newName\<rparr>"

primrec Replace:: "Component\<Rightarrow>Name\<Rightarrow>Component\<Rightarrow>Component" where
  "Replace (Primitive N itf s) N1 C = (Primitive N itf s)" |
  "Replace (Composite N itf sub bindings s) N1 C = 
      addSubCp (removeSubCp (Composite N itf sub
           ((\<lambda> b.(RenameBinding b N1 (getName C)))` bindings)  s) N1) C"

lemma removeEmpty:
"sub^N1 = None \<Longrightarrow>removeSubCp (Composite N itf sub bindings  s) N1 = Composite N itf sub bindings  s"
apply auto;
done;

lemma remove_diff_Name_CL:
" distinct (map getName sub)\<longrightarrow>C'\<in>set sub\<longrightarrow> C\<in>set (remove1 C' sub)\<longrightarrow>getName C'\<noteq>getName C"
apply (induct_tac sub);
apply auto;
apply (subgoal_tac "getName C' \<in> getName ` set list");
apply simp;
apply (rule_tac x=C' in image_eqI);
apply auto;
done;


lemma remove_diff_Name_Comp:
"\<lbrakk>CorrectComponentStructure (Composite N itf sub bindings  s);C'\<in>set sub; C\<in>set (remove1 C' sub)\<rbrakk>\<Longrightarrow>getName C'\<noteq>getName C"
apply (unfold CorrectComponentStructure_def);
apply (insert remove_diff_Name_CL);
apply auto;
done;

declare RenameBinding_def[simp];

lemma CompleteList_remove1[rule_format]: "CompleteList CL \<longrightarrow> CompleteList (remove1 C CL)"
apply (induct_tac CL);
apply auto;
done;


lemma replace_complete: 
" \<lbrakk>sub^(getName C')=None;sub^N'=Some oldC;getItfs oldC = getItfs C';Complete (Composite N itf sub bindings s); Complete C'; CorrectComponentStructure (Composite N itf sub bindings s);CorrectComponentStructure C'\<rbrakk>\<Longrightarrow> Complete (Replace (Composite N itf sub bindings s) N' C')"
apply auto;
apply (subgoal_tac  "remove1 oldC sub ^ getName C' =  None")
apply (frule getCp_inlist);
apply auto;
apply (simp add: allExternalItfsBound_def )
apply clarify;
apply (drule_tac x=oldC in bspec);
apply force;
apply (drule_tac x=IN in spec);
apply (drule_tac x=itype in spec);
apply simp+;

apply (case_tac "kind itype");
apply simp+;
apply (elim bexE);
apply (rule_tac x=b in bexI);
apply auto;
apply (rule_tac x=b in bexI);
apply auto;


apply (drule_tac x=C in bspec);
apply (subgoal_tac " set (remove1 oldC sub) \<subseteq> set sub");
apply force;
apply (rule List.set_remove1_subset);
apply (simp add: allExternalItfsBound_def )
apply clarify;
apply (drule_tac x=IN in spec);
apply (drule_tac x=itype in spec);
apply simp+;

apply (case_tac "kind itype");
apply simp+;
apply (elim bexE);
apply (rule_tac x=b in bexI);
apply auto;
apply (subgoal_tac "CorrectComponentStructure (Composite N itf sub bindings s)")
apply (drule remove_diff_Name_Comp);
apply auto;
apply (rule_tac x=b in bexI);
apply auto;
apply (subgoal_tac "CorrectComponentStructure (Composite N itf sub bindings s)")
apply (drule remove_diff_Name_Comp);
apply auto;
(*3*)
apply (simp add: allInternalItfsBound_def )
apply auto;

apply (drule_tac x=IN in spec);
apply (drule_tac x=itype in spec);
apply simp+;
apply (case_tac "kind itype");
apply simp+;
apply (elim bexE);
apply (rule_tac x=b in bexI);
apply auto;
apply (rule_tac x=b in bexI);
apply auto;

apply (rule CompleteList_remove1);
apply simp;

apply (rule getCp_None_remove1,simp);
done;


lemma replace_correct: 
" \<lbrakk>sub^(getName C')=None;sub^N'=Some oldC;getItfs oldC = getItfs C';Complete (Composite N itf sub bindings s); Complete C'; CorrectComponentStructure (Composite N itf sub bindings s);CorrectComponentStructure C'\<rbrakk>\<Longrightarrow> 
CorrectComponentStructure (Replace (Composite N itf sub bindings s) N' C')"
apply auto;
apply (subgoal_tac  "remove1 oldC sub ^ getName C' =  None")
apply (frule getCp_inlist);
apply auto;
apply (case_tac "src b");
apply auto;
apply (case_tac "dest b");
apply auto;
(*5*)
apply (simp add: NoDuplicateSrc_def);
apply auto;
apply (case_tac "src b1");
apply (case_tac "src b2");
apply auto;
(*6*)
apply (rotate_tac 8);
apply (drule_tac x=b1 in bspec)
apply simp;
apply (rotate_tac -1)
apply (drule_tac x=b2 in bspec)
apply simp+;
apply (case_tac "src b2");
apply (auto);
(*5*)
apply (rotate_tac 8);
apply (drule_tac x=b1 in bspec)
apply simp;
apply (rotate_tac -1)
apply (drule_tac x=b2 in bspec)
apply (auto split: split_if_asm);
(*4*)
apply (drule_tac C=oldC in getCp_None_remove1);
apply simp+;
apply (subgoal_tac "getName x = getName x");
apply (drule_tac C=x and N="getName x" in getCpI2);
apply auto;
apply (rule map_remove1)
apply simp;
(*2*)
apply (subgoal_tac " \<forall> C CL.
 CorrectComponentStructureList CL \<longrightarrow>CorrectComponentStructureList (remove1 C CL)")
apply force;
apply (intro allI)
apply (induct_tac CL)
apply auto
apply (rule getCp_None_remove1,simp);
done;
end